# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from dataclasses import dataclass
from types import TracebackType
from typing import Generator, TypeAlias

from dace import Memlet, data, dtypes, nodes

from gt4py import eve
from gt4py.cartesian.gtc import common, definitions
from gt4py.cartesian.gtc.dace import utils


SymbolDict: TypeAlias = dict[str, dtypes.typeclass]


@dataclass
class Context:
    root: TreeRoot
    current_scope: TreeScope

    field_extents: dict[str, definitions.Extent]  # field_name -> Extent
    block_extents: dict[int, definitions.Extent]  # id(horizontal execution) -> Extent


class ContextPushPop:
    """Append the node to the scope, then push/pop the scope."""

    def __init__(self, ctx: Context, node: TreeScope) -> None:
        self._ctx = ctx
        self._parent_scope = ctx.current_scope
        self._node = node

    def __enter__(self) -> None:
        self._node.parent = self._parent_scope
        self._parent_scope.children.append(self._node)
        self._ctx.current_scope = self._node

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._ctx.current_scope = self._parent_scope


class Axis(eve.StrEnum):
    I = "I"  # noqa: E741 [ambiguous-variable-name]
    J = "J"
    K = "K"

    def domain_symbol(self) -> eve.SymbolRef:
        return eve.SymbolRef(f"__{self.upper()}")

    def iteration_symbol(self) -> eve.SymbolRef:
        return eve.SymbolRef(f"__{self.lower()}")

    @staticmethod
    def dims_3d() -> Generator[Axis, None, None]:
        yield from [Axis.I, Axis.J, Axis.K]

    @staticmethod
    def dims_horizontal() -> Generator[Axis, None, None]:
        yield from [Axis.I, Axis.J]

    def to_idx(self) -> int:
        return [Axis.I, Axis.J, Axis.K].index(self)

    def domain_dace_symbol(self):
        return utils.get_dace_symbol(self.domain_symbol())

    def iteration_dace_symbol(self):
        return utils.get_dace_symbol(self.iteration_symbol())

    def tile_dace_symbol(self):
        return utils.get_dace_symbol(self.tile_symbol())


class Bounds(eve.Node):
    start: str
    end: str


class TreeNode(eve.Node):
    parent: TreeScope | None


class TreeScope(TreeNode):
    children: list[TreeScope | TreeNode]

    def scope(self, ctx: Context) -> ContextPushPop:
        return ContextPushPop(ctx, self)


class Tasklet(TreeNode):
    tasklet: nodes.Tasklet

    inputs: dict[str, Memlet]
    """Mapping tasklet.in_connectors to Memlets"""
    outputs: dict[str, Memlet]
    """Mapping tasklet.out_connectors to Memlets"""


class IfElse(TreeScope):
    # This should become an if/else, someday, so I am naming it if/else in hope
    # to see it before my bodily demise
    if_condition_code: str
    """Condition as ScheduleTree worthy code"""


class While(TreeScope):
    condition_code: str
    """Condition as ScheduleTree worthy code"""


class HorizontalLoop(TreeScope):
    bounds_i: Bounds
    bounds_j: Bounds

    schedule: dtypes.ScheduleType


class VerticalLoop(TreeScope):
    loop_order: common.LoopOrder
    bounds_k: Bounds

    schedule: dtypes.ScheduleType


class TreeRoot(TreeScope):
    name: str

    containers: dict[str, data.Data]
    """Mapping field/scalar names to data descriptors."""

    dimensions: dict[str, tuple[bool, bool, bool]]
    """Mapping field names to shape-axis."""

    shift: dict[str, dict[Axis, int]]
    """Mapping field names to dict[axis] -> shift."""

    symbols: SymbolDict
    """Mapping between type and symbol name."""



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import re
from typing import Final, Literal, Mapping, Optional, Union

import dace

from gt4py.next import common as gtx_common
from gt4py.next.type_system import type_specifications as ts


# arrays for connectivity tables use the following prefix
CONNECTIVITY_INDENTIFIER_PREFIX: Final[str] = "gt_conn_"
CONNECTIVITY_INDENTIFIER_RE: Final[re.Pattern] = re.compile(r"^gt_conn_(.+)$")

# regex for domain range symbols
RANGE_SYMBOL_RE: Final[re.Pattern] = re.compile(r"^__(.+)_\d+_range_[01]$")

# regex for field stride symbols
SIZE_SYMBOL_RE: Final[re.Pattern] = re.compile(r"^__(.+)_size_\d+$")

# regex for field stride symbols
STRIDE_SYMBOL_RE: Final[re.Pattern] = re.compile(r"^__(.+)_stride_\d+$")


def as_dace_type(type_: ts.ScalarType) -> dace.typeclass:
    """Converts GT4Py scalar type to corresponding DaCe type."""

    match type_.kind:
        case ts.ScalarKind.BOOL:
            return dace.bool_
        case ts.ScalarKind():
            return getattr(dace, type_.kind.name.lower())
        case _:
            raise ValueError(f"Scalar type '{type_}' not supported.")


def as_itir_type(dtype: dace.typeclass) -> ts.ScalarType:
    """Get GT4Py scalar representation of a DaCe type."""
    type_name = str(dtype.as_numpy_dtype())
    try:
        kind = getattr(ts.ScalarKind, type_name.upper())
    except AttributeError as ex:
        raise ValueError(f"Data type {type_name} not supported.") from ex
    return ts.ScalarType(kind)


def connectivity_identifier(name: str) -> str:
    return f"{CONNECTIVITY_INDENTIFIER_PREFIX}{name}"


def is_connectivity_identifier(
    name: str, offset_provider_type: Optional[gtx_common.OffsetProviderType] = None
) -> bool:
    m = CONNECTIVITY_INDENTIFIER_RE.match(name)
    if m is None:
        return False
    if offset_provider_type is None:
        # If no offset provider type is provided, we assume there is a connectivity identifier
        # that matches the CONNECTIVITY_INDENTIFIER_RE.
        return True
    return m[1] in offset_provider_type


def is_connectivity_symbol(name: str, offset_provider_type: gtx_common.OffsetProviderType) -> bool:
    m = SIZE_SYMBOL_RE.match(name) or STRIDE_SYMBOL_RE.match(name)
    if m is None:
        return False
    m = CONNECTIVITY_INDENTIFIER_RE.match(m[1])
    if m is None:
        return False
    return m[1] in offset_provider_type


def field_symbol_name(field_name: str, axis: int, sym: Literal["size", "stride"]) -> str:
    return f"__{field_name}_{sym}_{axis}"


def field_size_symbol_name(field_name: str, axis: int) -> str:
    return field_symbol_name(field_name, axis, "size")


def field_stride_symbol_name(field_name: str, axis: int) -> str:
    return field_symbol_name(field_name, axis, "stride")


def range_start_symbol(field_name: str, axis: int) -> str:
    """Format name of start symbol for domain range, as expected by GTIR."""
    return f"__{field_name}_{axis}_range_0"


def range_stop_symbol(field_name: str, axis: int) -> str:
    """Format name of stop symbol for domain range, as expected by GTIR."""
    return f"__{field_name}_{axis}_range_1"


def is_range_symbol(name: str) -> bool:
    return RANGE_SYMBOL_RE.match(name) is not None


def is_size_symbol(name: str) -> bool:
    return SIZE_SYMBOL_RE.match(name) is not None


def is_stride_symbol(name: str) -> bool:
    return STRIDE_SYMBOL_RE.match(name) is not None


def filter_connectivity_types(
    offset_provider_type: gtx_common.OffsetProviderType,
) -> dict[str, gtx_common.NeighborConnectivityType]:
    """
    Filter offset provider types of type `NeighborConnectivityType`.

    In other words, filter out the cartesian offset providers.
    """
    return {
        offset: conn
        for offset, conn in offset_provider_type.items()
        if isinstance(conn, gtx_common.NeighborConnectivityType)
    }


def safe_replace_symbolic(
    val: dace.symbolic.SymbolicType,
    symbol_mapping: Mapping[
        Union[dace.symbolic.SymbolicType, str], Union[dace.symbolic.SymbolicType, str]
    ],
) -> dace.symbolic.SymbolicType:
    """
    Replace free symbols in a dace symbolic expression, using `safe_replace()`
    in order to avoid clashes in case the new symbol value is also a free symbol
    in the original exoression.

    Args:
        val: The symbolic expression where to apply the replacement.
        symbol_mapping: The mapping table for symbol replacement.

    Returns:
        A new symbolic expression as result of symbol replacement.
    """
    # The list `x` is needed because `subs()` returns a new object and can not handle
    # replacement dicts of the form `{'x': 'y', 'y': 'x'}`.
    # The utility `safe_replace()` will call `subs()` twice in case of such dicts.
    x = [val]
    dace.symbolic.safe_replace(symbol_mapping, lambda m, xx=x: xx.append(xx[-1].subs(m)))
    return x[-1]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
import dataclasses
import functools
import typing
from collections.abc import MutableMapping
from typing import Any, Callable, Generic, Protocol, TypeVar

from typing_extensions import Self


StartT = TypeVar("StartT")
StartT_contra = TypeVar("StartT_contra", contravariant=True)
EndT = TypeVar("EndT")
EndT_co = TypeVar("EndT_co", covariant=True)
NewEndT = TypeVar("NewEndT")
IntermediateT = TypeVar("IntermediateT")
HashT = TypeVar("HashT")
DataT = TypeVar("DataT")
ArgT = TypeVar("ArgT")


def make_step(function: Workflow[StartT, EndT]) -> ChainableWorkflowMixin[StartT, EndT]:
    """
    Wrap a function in the workflow step convenience wrapper.

    Examples:
    ---------
    >>> @make_step
    ... def times_two(x: int) -> int:
    ...     return x * 2

    >>> def stringify(x: int) -> str:
    ...     return str(x)

    >>> # create a workflow int -> int -> str
    >>> times_two.chain(stringify)(3)
    '6'
    """
    return StepSequence.start(function)


@typing.runtime_checkable
class Workflow(Protocol[StartT_contra, EndT_co]):
    """
    Workflow protocol.

    Anything that implements this interface can be a workflow of one or more steps.
    - callable
    - take a single input argument
    """

    def __call__(self, inp: StartT_contra) -> EndT_co: ...


class ReplaceEnabledWorkflowMixin(Workflow[StartT_contra, EndT_co], Protocol):
    """
    Subworkflow replacement mixin.

    Any subclass MUST be a dataclass for `.replace` to work
    """

    def replace(self, **kwargs: Any) -> Self:
        """
        Build a new instance with replaced substeps.

        Raises:
            TypeError: If `self` is not a dataclass.
        """
        if not dataclasses.is_dataclass(self):
            raise TypeError(f"'{self.__class__}' is not a dataclass.")
        assert not isinstance(self, type)
        return dataclasses.replace(self, **kwargs)


class ChainableWorkflowMixin(Workflow[StartT, EndT_co], Protocol[StartT, EndT_co]):
    def chain(
        self, next_step: Workflow[EndT_co, NewEndT]
    ) -> ChainableWorkflowMixin[StartT, NewEndT]:
        return make_step(self).chain(next_step)


@dataclasses.dataclass(frozen=True)
class NamedStepSequence(
    ChainableWorkflowMixin[StartT, EndT], ReplaceEnabledWorkflowMixin[StartT, EndT]
):
    """
    Workflow with linear succession of named steps.

    Examples:
    ---------
    >>> import dataclasses

    >>> def parse(x: str) -> int:
    ...     return int(x)

    >>> def plus_half(x: int) -> float:
    ...     return x + 0.5

    >>> def stringify(x: float) -> str:
    ...     return str(x)

    >>> @dataclasses.dataclass(frozen=True)
    ... class ParseOpPrint(NamedStepSequence[str, str]):
    ...     parse: Workflow[str, int]
    ...     op: Workflow[int, float]
    ...     print: Workflow[float, str]

    >>> pop = ParseOpPrint(parse=parse, op=plus_half, print=stringify)

    >>> pop.step_order
    ['parse', 'op', 'print']

    >>> pop(73)
    '73.5'

    >>> def plus_tenth(x: int) -> float:
    ...     return x + 0.1


    >>> pop.replace(op=plus_tenth)(73)
    '73.1'
    """

    def __call__(self, inp: StartT) -> EndT:
        """Compose the steps in the order defined in the `.step_order` class attribute."""
        step_result: Any = inp
        for step_name in self.step_order:
            step_result = getattr(self, step_name)(step_result)
        return step_result

    @functools.cached_property
    def step_order(self) -> list[str]:
        """
        Read step order from class definition by default.

        Only attributes who are type hinted to be of a type that
        conforms to the Workflow protocol are considered steps.
        """
        step_names: list[str] = []
        annotations = typing.get_type_hints(self.__class__)
        for field in dataclasses.fields(self):
            field_type = annotations[field.name]
            field_type = typing.get_origin(field_type) or field_type
            if issubclass(field_type, Workflow):
                step_names.append(field.name)
        return step_names


@dataclasses.dataclass(frozen=True)
class MultiWorkflow(
    ChainableWorkflowMixin[StartT, EndT], ReplaceEnabledWorkflowMixin[StartT, EndT]
):
    """A flexible workflow, where the sequence of steps depends on the input type."""

    def __call__(self, inp: StartT) -> EndT:
        step_result: Any = inp
        for step_name in self.step_order(inp):
            step_result = getattr(self, step_name)(step_result)
        return step_result

    @abc.abstractmethod
    def step_order(self, inp: StartT) -> list[str]:
        pass


@dataclasses.dataclass(frozen=True)
class StepSequence(ChainableWorkflowMixin[StartT, EndT]):
    """
    Composable workflow of single input callables.

    Examples:
    ---------
    >>> def plus_one(x: int) -> int:
    ...     return x + 1

    >>> def plus_half(x: int) -> float:
    ...     return x + 0.5

    >>> def stringify(x: float) -> str:
    ...     return str(x)

    >>> StepSequence.start(plus_one).chain(plus_half).chain(stringify)(73)
    '74.5'

    """

    @dataclasses.dataclass(frozen=True)
    class __Steps:
        inner: tuple[Workflow[Any, Any], ...]

    # todo(ricoh): replace with normal tuple with TypeVarTuple hints
    #   to enable automatic deduction StartT and EndT fom constructor
    #   calls. TypeVarTuple is available in typing_extensions in
    #   Python <= 3.11. Revise after mypy constraint is > 1.0.1,
    #   which fails on trying to check TypeVarTuple.
    steps: __Steps

    def __call__(self, inp: StartT) -> EndT:
        step_result: Any = inp
        for step in self.steps.inner:
            step_result = step(step_result)
        return step_result

    def chain(self, next_step: Workflow[EndT, NewEndT]) -> ChainableWorkflowMixin[StartT, NewEndT]:
        return typing.cast(
            ChainableWorkflowMixin[StartT, NewEndT],
            self.__class__(self.__Steps((*self.steps.inner, next_step))),
        )

    @classmethod
    def start(cls, first_step: Workflow[StartT, EndT]) -> ChainableWorkflowMixin[StartT, EndT]:
        return cls(cls.__Steps((first_step,)))


@dataclasses.dataclass(frozen=True)
class CachedStep(
    ChainableWorkflowMixin[StartT, EndT],
    ReplaceEnabledWorkflowMixin[StartT, EndT],
    Generic[StartT, EndT, HashT],
):
    """
    Cached workflow of single input callables.

    Examples:
    ---------
    >>> def heavy_computation(x: int) -> int:
    ...     print("This might take a while...")
    ...     return x

    >>> cached_step = CachedStep(step=heavy_computation)

    >>> cached_step(42)
    This might take a while...
    42

    The next invocation for the same argument will be cached:
    >>> cached_step(42)
    42

    >>> cached_step(1)
    This might take a while...
    1
    """

    step: Workflow[StartT, EndT]
    hash_function: Callable[[StartT], HashT] = dataclasses.field(default=hash)  # type: ignore[assignment]
    cache: MutableMapping[HashT, EndT] = dataclasses.field(repr=False, default_factory=dict)

    def __call__(self, inp: StartT) -> EndT:
        """Run the step only if the input is not cached, else return from cache."""
        hash_ = self.hash_function(inp)
        try:
            result = self.cache[hash_]
        except KeyError:
            result = self.cache[hash_] = self.step(inp)
        return result


@dataclasses.dataclass(frozen=True)
class SkippableStep(
    ChainableWorkflowMixin[StartT, EndT], ReplaceEnabledWorkflowMixin[StartT, EndT]
):
    step: Workflow[StartT, EndT]

    def __call__(self, inp: StartT) -> EndT:
        if not self.skip_condition(inp):
            return self.step(inp)
        return inp  # type: ignore[return-value]  # up to the implementer to make sure StartT == EndT

    def skip_condition(self, inp: StartT) -> bool:
        raise NotImplementedError()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Any, Dict, Iterable, Iterator, List, Optional, TypeGuard, Union

import gt4py.eve as eve
from gt4py.eve import NodeTranslator, concepts
from gt4py.eve.utils import UIDGenerator
from gt4py.next.program_processors.codegens.gtfn import gtfn_ir, gtfn_ir_common
from gt4py.next.program_processors.codegens.gtfn.gtfn_im_ir import (
    AssignStmt,
    Conditional,
    EmptyListInitializer,
    ImperativeFunctionDefinition,
    InitStmt,
    ReturnStmt,
    Stmt,
)


# TODO: start of code clone from unroll_reduce.py. This is necessary since whilet the IR nodes are compatible between itir and gtfn_ir,
#       the structure of the ir is slightly different, hence functions like _is_shifted and _get_partial_offset_tag are slightly changed
#       in this version of the code clone. To be removed asap
def _is_shifted(arg: gtfn_ir_common.Expr) -> TypeGuard[gtfn_ir.FunCall]:
    return (
        isinstance(arg, gtfn_ir.FunCall)
        and isinstance(arg.fun, gtfn_ir_common.SymRef)
        and arg.fun.id == "shift"
    )


def _is_applied_lift(arg: gtfn_ir_common.Expr) -> TypeGuard[gtfn_ir.FunCall]:
    return (
        isinstance(arg, gtfn_ir.FunCall)
        and isinstance(arg.fun, gtfn_ir.FunCall)
        and arg.fun.fun == gtfn_ir_common.SymRef(id="lift")
    )


def _is_shifted_or_lifted_and_shifted(arg: gtfn_ir_common.Expr) -> TypeGuard[gtfn_ir.FunCall]:
    return _is_shifted(arg) or (
        _is_applied_lift(arg)
        and any(_is_shifted_or_lifted_and_shifted(nested_arg) for nested_arg in arg.args)
    )


def _get_shifted_args(reduce_args: Iterable[gtfn_ir_common.Expr]) -> Iterator[gtfn_ir.FunCall]:
    return filter(_is_shifted_or_lifted_and_shifted, reduce_args)


def _is_list_of_funcalls(lst: list) -> TypeGuard[list[gtfn_ir.FunCall]]:
    return all(isinstance(f, gtfn_ir.FunCall) for f in lst)


def _get_partial_offset_tag(arg: gtfn_ir.FunCall) -> str:
    if _is_shifted(arg):
        assert isinstance(arg.fun, gtfn_ir_common.SymRef)
        offset = arg.args[-1]
        assert isinstance(offset, gtfn_ir.OffsetLiteral)
        assert isinstance(offset.value, str)
        return offset.value
    else:
        assert _is_applied_lift(arg)
        assert _is_list_of_funcalls(arg.args)
        partial_offsets = [_get_partial_offset_tag(arg) for arg in arg.args]
        assert all(o == partial_offsets[0] for o in partial_offsets)
        return partial_offsets[0]


def _get_partial_offset_tags(reduce_args: Iterable[gtfn_ir_common.Expr]) -> Iterable[str]:
    return [_get_partial_offset_tag(arg) for arg in _get_shifted_args(reduce_args)]


def _is_reduce(node: gtfn_ir.FunCall) -> TypeGuard[gtfn_ir.FunCall]:
    return isinstance(node.fun, gtfn_ir.FunCall) and node.fun.fun == gtfn_ir_common.SymRef(
        id="reduce"
    )


# TODO: end of code clone


class PlugInCurrentIdx(NodeTranslator):
    def visit_SymRef(
        self, node: gtfn_ir_common.SymRef
    ) -> gtfn_ir.OffsetLiteral | gtfn_ir_common.SymRef:
        if node.id == "nbh_iter":
            return self.cur_idx
        if self.acc is not None and node.id == self.acc.id:
            return gtfn_ir_common.SymRef(id=self.red_idx)
        return self.generic_visit(node)

    def __init__(
        self, cur_idx: gtfn_ir.OffsetLiteral, acc: Optional[gtfn_ir_common.Sym], red_idx: str
    ):
        self.cur_idx = cur_idx
        self.acc = acc
        self.red_idx = red_idx


@dataclasses.dataclass(frozen=False)
class GTFN_IM_lowering(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    # we use one UID generator per instance such that the generated ids are
    # stable across multiple runs (required for caching to properly work)
    uids: UIDGenerator = dataclasses.field(init=False, repr=False, default_factory=UIDGenerator)

    def visit_SymRef(self, node: gtfn_ir_common.SymRef, **kwargs: Any) -> gtfn_ir_common.SymRef:
        if "localized_symbols" in kwargs and node.id in kwargs["localized_symbols"]:
            return gtfn_ir_common.SymRef(id=kwargs["localized_symbols"][node.id])
        return node

    def commit_args(
        self, node: gtfn_ir.FunCall, tmp_id: str, fun_id: str, **kwargs: Any
    ) -> gtfn_ir.FunCall:
        for i, arg in enumerate(node.args):
            expr = self.visit(arg, **kwargs)
            self.imp_list_ir.append(InitStmt(lhs=gtfn_ir_common.Sym(id=f"{tmp_id}_{i}"), rhs=expr))
        tup_args = [gtfn_ir_common.SymRef(id=f"{tmp_id}_{i}") for i in range(len(node.args))]
        return gtfn_ir.FunCall(fun=gtfn_ir_common.SymRef(id=fun_id), args=tup_args)

    def _expand_lambda(
        self,
        node: gtfn_ir.FunCall,
        new_args: List[gtfn_ir.FunCall],
        red_idx: str,
        max_neighbors: int,
        **kwargs: Any,
    ) -> None:
        fun, init = node.fun.args  # type: ignore
        param_to_args = dict(zip([param.id for param in fun.params[1:]], new_args))
        acc = fun.params[0]

        class InlineArgs(NodeTranslator):
            def visit_Expr(self, node: gtfn_ir_common.Expr) -> gtfn_ir_common.Expr:
                if hasattr(node, "id") and node.id in param_to_args:
                    return param_to_args[node.id]
                return self.generic_visit(node)

        new_body = InlineArgs().visit(fun.expr)

        red_lit = gtfn_ir_common.Sym(id=f"{red_idx}")
        self.imp_list_ir.append(InitStmt(lhs=red_lit, rhs=self.visit(init, **kwargs)))

        for i in range(max_neighbors):
            new_expr = PlugInCurrentIdx(
                cur_idx=gtfn_ir.OffsetLiteral(value=i), acc=acc, red_idx=red_idx
            ).visit(new_body)
            rhs = self.visit(new_expr, **kwargs)
            self.imp_list_ir.append(AssignStmt(lhs=gtfn_ir_common.SymRef(id=red_idx), rhs=rhs))

    def _expand_symref(
        self,
        node: gtfn_ir.FunCall,
        new_args: List[gtfn_ir.FunCall],
        red_idx: str,
        max_neighbors: int,
        **kwargs: Any,
    ) -> None:
        fun, init = node.fun.args  # type: ignore

        red_lit = gtfn_ir_common.Sym(id=f"{red_idx}")
        self.imp_list_ir.append(InitStmt(lhs=red_lit, rhs=self.visit(init, **kwargs)))

        for i in range(max_neighbors):
            plugged_in_args = [
                PlugInCurrentIdx(
                    cur_idx=gtfn_ir.OffsetLiteral(value=i), acc=None, red_idx=red_idx
                ).visit(arg)
                for arg in new_args
            ]
            rhs = gtfn_ir.FunCall(
                fun=fun, args=[gtfn_ir_common.SymRef(id=red_idx), *plugged_in_args]
            )
            self.imp_list_ir.append(AssignStmt(lhs=gtfn_ir_common.SymRef(id=red_idx), rhs=rhs))

    def visit_FunCall(self, node: gtfn_ir.FunCall, **kwargs: Any) -> gtfn_ir_common.Expr:
        if any(isinstance(arg, gtfn_ir.Lambda) for arg in node.args):
            # do not try to lower constructs that take lambdas as argument to something more readable
            lam_idx = self.uids.sequential_id(prefix="lam")
            self.imp_list_ir.append(InitStmt(lhs=gtfn_ir_common.Sym(id=f"{lam_idx}"), rhs=node))
            return gtfn_ir_common.SymRef(id=f"{lam_idx}")
        if isinstance(node.fun, gtfn_ir.Lambda):
            localized_symbols = {**kwargs["localized_symbols"]}  # create a new scope

            lam_idx = self.uids.sequential_id(prefix="lam")
            params = [self.visit(param, **kwargs) for param in node.fun.params]
            args = [self.visit(arg, **kwargs) for arg in node.args]
            for param, arg in zip(params, args):
                localized_symbols[param.id] = new_symbol = (
                    f"{param.id}_{self.uids.sequential_id()}_local"
                )
                self.imp_list_ir.append(
                    InitStmt(
                        lhs=gtfn_ir_common.Sym(id=new_symbol),
                        rhs=arg,
                    )
                )
            new_kwargs = {**kwargs, "localized_symbols": localized_symbols}
            expr = self.visit(node.fun.expr, **new_kwargs)
            self.imp_list_ir.append(InitStmt(lhs=gtfn_ir_common.Sym(id=f"{lam_idx}"), rhs=expr))
            return gtfn_ir_common.SymRef(id=f"{lam_idx}")
        if _is_reduce(node):
            raise AssertionError(
                "Not implemented. The code-path was removed as it was not actively used and tested."
            )
        if isinstance(node.fun, gtfn_ir_common.SymRef) and node.fun.id == "make_tuple":
            tupl_id = self.uids.sequential_id(prefix="tupl")
            tuple_fun = self.commit_args(node, tupl_id, "make_tuple", **kwargs)
            self.imp_list_ir.append(
                InitStmt(lhs=gtfn_ir_common.Sym(id=f"{tupl_id}"), rhs=tuple_fun)
            )
            return gtfn_ir_common.SymRef(id=f"{tupl_id}")
        return gtfn_ir.FunCall(
            fun=self.visit(node.fun, **kwargs),
            args=[self.visit(arg, **kwargs) for arg in node.args],
        )

    def visit_TernaryExpr(self, node: gtfn_ir.TernaryExpr, **kwargs: Any) -> gtfn_ir_common.SymRef:
        cond = self.visit(node.cond, **kwargs)
        if_ = self.visit(node.true_expr, **kwargs)
        else_ = self.visit(node.false_expr, **kwargs)
        cond_idx = self.uids.sequential_id(prefix="cond")
        self.imp_list_ir.append(
            Conditional(
                cond_type=f"{cond_idx}_t",
                init_stmt=InitStmt(
                    init_type=f"{cond_idx}_t",
                    lhs=gtfn_ir_common.Sym(id=cond_idx),
                    rhs=EmptyListInitializer(),
                ),
                cond=cond,
                if_stmt=AssignStmt(lhs=gtfn_ir_common.SymRef(id=cond_idx), rhs=if_),
                else_stmt=AssignStmt(lhs=gtfn_ir_common.SymRef(id=cond_idx), rhs=else_),
            )
        )
        return gtfn_ir_common.SymRef(id=cond_idx)

    def visit_FunctionDefinition(
        self, node: gtfn_ir.FunctionDefinition, **kwargs: Any
    ) -> ImperativeFunctionDefinition:
        self.imp_list_ir: List[Union[Stmt, Conditional]] = []
        self.sym_table: Dict[gtfn_ir_common.Sym, gtfn_ir_common.SymRef] = node.annex.symtable
        ret = self.visit(node.expr, localized_symbols={}, **kwargs)

        return ImperativeFunctionDefinition(
            id=node.id, params=node.params, fun=[*self.imp_list_ir, ReturnStmt(ret=ret)]
        )

    def visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
        kwargs = {
            **kwargs,
            "inside_fun": kwargs.get("inside_fun", False)
            or isinstance(node, gtfn_ir.FunctionDefinition),
        }

        # only transform things inside of function definitions and leave the rest as is
        if kwargs["inside_fun"]:
            # we are inside a function, visit as usual
            return super().visit(node, **kwargs)
        else:
            # we are outside of a function, don't transform anything (i.e. don't call any `vist_`
            # methods on `node`), but just visit child nodes.
            return self.generic_visit(node, **kwargs)

    def visit_ScanPassDefinition(
        self, node: gtfn_ir.ScanPassDefinition, **kwargs: Any
    ) -> gtfn_ir.ScanPassDefinition:
        return node

    def visit_ScanExecution(
        self, node: gtfn_ir.ScanExecution, **kwargs: Any
    ) -> gtfn_ir.ScanExecution:
        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Literal

from gt4py.next import common
from gt4py.next.type_system import type_specifications as ts


class NamedRangeType(ts.TypeSpec):
    dim: common.Dimension


class OffsetLiteralType(ts.TypeSpec):
    value: ts.ScalarType | str


class IteratorType(ts.DataType, ts.CallableType):
    position_dims: list[common.Dimension] | Literal["unknown"]
    defined_dims: list[common.Dimension]
    element_type: ts.DataType


class ProgramType(ts.TypeSpec):
    params: dict[str, ts.DataType]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import importlib
import os
import pathlib
from typing import Any, Callable, Sequence

import dace
import factory

from gt4py._core import definitions as core_defs, locking
from gt4py.next import config
from gt4py.next.otf import languages, stages, step_types, workflow
from gt4py.next.program_processors.runners.dace.workflow import common as gtx_wfcommon


def _get_sdfg_ctype_arglist_callback(
    module_name: str, bind_func_name: str, python_code: str
) -> Callable[
    [core_defs.DeviceType, Sequence[dace.dtypes.Data], Sequence[Any], Sequence[Any]], None
]:
    """
    Helper method to load dynamically generated Python code as a module and return
    a function to update the list of SDFG call arguments.

    Args:
        module_name: Name to use to load the python code as a module.
        bind_func_name: Name to use for the translation function.
        python_code: String containg the Python code to load.

    Returns:
        A callable object to update the list of SDFG call arguments.
    """
    spec = importlib.util.spec_from_loader(module_name, loader=None)
    assert spec is not None
    module = importlib.util.module_from_spec(spec)
    exec(python_code, module.__dict__)
    assert bind_func_name in module.__dict__
    return module.__dict__[bind_func_name]


class CompiledDaceProgram(stages.ExtendedCompiledProgram):
    sdfg_program: dace.CompiledSDFG

    # Sorted list of SDFG arguments as they appear in program ABI and corresponding data type;
    # scalar arguments that are not used in the SDFG will not be present.
    sdfg_argtypes: list[dace.dtypes.Data]

    # The compiled program contains a callable object to update the SDFG arguments list.
    update_sdfg_ctype_arglist: Callable[
        [core_defs.DeviceType, Sequence[dace.dtypes.Data], Sequence[Any], Sequence[Any]],
        None,
    ]

    def __init__(
        self,
        program: dace.CompiledSDFG,
        bind_func_name: str,
        binding_source: stages.BindingSource[languages.SDFG, languages.Python],
        implicit_domain: bool,
    ):
        self.sdfg_program = program
        self.implicit_domain = implicit_domain

        # `dace.CompiledSDFG.arglist()` returns an ordered dictionary that maps the argument
        # name to its data type, in the same order as arguments appear in the program ABI.
        # This is also the same order of arguments in `dace.CompiledSDFG._lastargs[0]`.
        self.sdfg_argtypes = list(program.sdfg.arglist().values())

        # Note that `binding_source` contains Python code tailored to this specific SDFG.
        #   We need to ensure that it is loaded as a Python module with a unique name,
        #   in order to avoid conflicts with other variants of the same program.
        #   Therefore, we use the name of the build folder as module name.
        binding_module_name = os.path.basename(program.sdfg.build_folder)
        self.update_sdfg_ctype_arglist = _get_sdfg_ctype_arglist_callback(
            binding_module_name, bind_func_name, binding_source.source_code
        )

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        result = self.sdfg_program(*args, **kwargs)
        assert result is None

    def fast_call(self) -> None:
        result = self.sdfg_program.fast_call(*self.sdfg_program._lastargs)
        assert result is None


@dataclasses.dataclass(frozen=True)
class DaCeCompiler(
    workflow.ChainableWorkflowMixin[
        stages.CompilableSource[languages.SDFG, languages.LanguageSettings, languages.Python],
        CompiledDaceProgram,
    ],
    workflow.ReplaceEnabledWorkflowMixin[
        stages.CompilableSource[languages.SDFG, languages.LanguageSettings, languages.Python],
        CompiledDaceProgram,
    ],
    step_types.CompilationStep[languages.SDFG, languages.LanguageSettings, languages.Python],
):
    """Use the dace build system to compile a GT4Py program to a ``gt4py.next.otf.stages.CompiledProgram``."""

    bind_func_name: str
    cache_lifetime: config.BuildCacheLifetime
    device_type: core_defs.DeviceType
    cmake_build_type: config.CMakeBuildType = config.CMakeBuildType.DEBUG

    def __call__(
        self,
        inp: stages.CompilableSource[languages.SDFG, languages.LanguageSettings, languages.Python],
    ) -> CompiledDaceProgram:
        # NOTE: In an ideal world we should create a configuration context here. But,
        #   because of [DaCe issue#2125](https://github.com/spcl/dace/issues/2125) this
        #   is not possible and we have to set the configuration in full.
        # NOTE: We hope here that the build type does not varies across programs.
        gtx_wfcommon.set_dace_config(
            device_type=self.device_type,
            cmake_build_type=self.cmake_build_type,
        )
        sdfg = dace.SDFG.from_json(inp.program_source.source_code)
        sdfg_build_folder = pathlib.Path(sdfg.build_folder)
        sdfg_build_folder.mkdir(parents=True, exist_ok=True)
        with locking.lock(sdfg_build_folder):
            sdfg_program = sdfg.compile(validate=False)

        assert inp.binding_source is not None
        return CompiledDaceProgram(
            sdfg_program,
            self.bind_func_name,
            inp.binding_source,
            inp.program_source.implicit_domain,
        )


class DaCeCompilationStepFactory(factory.Factory):
    class Meta:
        model = DaCeCompiler



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, Optional, Set

from gt4py.eve import NodeTranslator, PreserveLocationVisitor, SymbolTableTrait
from gt4py.next.iterator import ir
from gt4py.next.iterator.type_system import inference as type_inference


class RemapSymbolRefs(PreserveLocationVisitor, NodeTranslator):
    # This pass preserves, but doesn't use the `type`, `recorded_shifts`, `domain` annex.
    PRESERVED_ANNEX_ATTRS = ("type", "recorded_shifts", "domain")

    def visit_SymRef(self, node: ir.SymRef, *, symbol_map: Dict[str, ir.Node]):
        return symbol_map.get(str(node.id), node)

    def visit_Lambda(self, node: ir.Lambda, *, symbol_map: Dict[str, ir.Node]):
        params = {str(p.id) for p in node.params}
        new_symbol_map = {k: v for k, v in symbol_map.items() if k not in params}
        return ir.Lambda(params=node.params, expr=self.visit(node.expr, symbol_map=new_symbol_map))

    def generic_visit(self, node: ir.Node, **kwargs: Any):  # type: ignore[override]
        assert isinstance(node, SymbolTableTrait) == isinstance(node, ir.Lambda), (
            "found unexpected new symbol scope"
        )
        return super().generic_visit(node, **kwargs)


class RenameSymbols(PreserveLocationVisitor, NodeTranslator):
    # This pass preserves, but doesn't use the `type`, `recorded_shifts`, `domain` annex.
    PRESERVED_ANNEX_ATTRS = ("type", "recorded_shifts", "domain")

    def visit_Sym(
        self, node: ir.Sym, *, name_map: Dict[str, str], active: Optional[Set[str]] = None
    ):
        if active and node.id in active:
            return ir.Sym(id=name_map.get(node.id, node.id))
        return node

    def visit_SymRef(
        self, node: ir.SymRef, *, name_map: Dict[str, str], active: Optional[Set[str]] = None
    ):
        if active and node.id in active:
            new_ref = ir.SymRef(id=name_map.get(node.id, node.id))
            type_inference.copy_type(from_=node, to=new_ref, allow_untyped=True)
            return new_ref
        return node

    def generic_visit(  # type: ignore[override]
        self, node: ir.Node, *, name_map: Dict[str, str], active: Optional[Set[str]] = None
    ):
        if isinstance(node, SymbolTableTrait):
            if active is None:
                active = set()
            active = active | set(node.annex.symtable)
        return super().generic_visit(node, name_map=name_map, active=active)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import enum
import json
import pathlib
from typing import Final, Optional


_DATAFILE_NAME: Final = "gt4py.json"


class BuildStatus(enum.IntEnum):
    INITIALIZED = enum.auto()
    CONFIGURED = enum.auto()
    COMPILED = enum.auto()


@dataclasses.dataclass(frozen=True)
class BuildData:
    """
    Tracks build status and results of GT4Py program compilations.

    Build system projects should keep this information up to date for each build,
    stored in a JSON file in the build folder at all times.

    The `module` property should only ever be used for compilation results which can
    directly be imported from python and contain a callable attribute named
    `entry_point_name` that represents a GT4Py program.
    """

    status: BuildStatus
    module: pathlib.Path
    entry_point_name: str

    def to_json(self) -> dict[str, str]:
        return {
            "status": self.status.name,
            "module": str(self.module),
            "entry_point_name": self.entry_point_name,
        }

    @classmethod
    def _status_to_json(cls, status: BuildStatus) -> str:
        return status.name

    @classmethod
    def _status_from_json(cls, status: str) -> BuildStatus:
        return getattr(BuildStatus, status)

    @classmethod
    def _module_from_json(cls, module: str) -> pathlib.Path:
        return pathlib.Path(module)

    @classmethod
    def from_json(cls, json_dict: dict[str, str]) -> BuildData:
        return cls(
            status=getattr(BuildStatus, json_dict["status"]),
            module=pathlib.Path(json_dict["module"]),
            entry_point_name=json_dict["entry_point_name"],
        )


def contains_data(path: pathlib.Path) -> bool:
    return (path / _DATAFILE_NAME).exists()


def read_data(path: pathlib.Path) -> Optional[BuildData]:
    try:
        return BuildData.from_json(json.loads((path / _DATAFILE_NAME).read_text()))
    except FileNotFoundError:
        return None


def write_data(data: BuildData, path: pathlib.Path) -> None:
    (path / _DATAFILE_NAME).write_text(json.dumps(data.to_json()))


def update_status(new_status: BuildStatus, path: pathlib.Path) -> None:
    old_data = read_data(path)
    assert old_data
    write_data(
        BuildData(
            status=new_status, module=old_data.module, entry_point_name=old_data.entry_point_name
        ),
        path,
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
import functools
from typing import Any, Final, Optional

import factory
import numpy as np

from gt4py._core import definitions as core_defs
from gt4py.eve import codegen
from gt4py.next import common
from gt4py.next.ffront import fbuiltins
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.transforms import pass_manager
from gt4py.next.otf import arguments, languages, stages, step_types, workflow
from gt4py.next.otf.binding import cpp_interface, interface
from gt4py.next.program_processors.codegens.gtfn.codegen import GTFNCodegen, GTFNIMCodegen
from gt4py.next.program_processors.codegens.gtfn.gtfn_ir_to_gtfn_im_ir import GTFN_IM_lowering
from gt4py.next.program_processors.codegens.gtfn.itir_to_gtfn_ir import GTFN_lowering
from gt4py.next.type_system import type_specifications as ts, type_translation


GENERATED_CONNECTIVITY_PARAM_PREFIX = "gt_conn_"


def get_param_description(name: str, type_: Any) -> interface.Parameter:
    return interface.Parameter(name, type_)


@dataclasses.dataclass(frozen=True)
class GTFNTranslationStep(
    workflow.ReplaceEnabledWorkflowMixin[
        stages.CompilableProgram,
        stages.ProgramSource[languages.NanobindSrcL, languages.LanguageWithHeaderFilesSettings],
    ],
    workflow.ChainableWorkflowMixin[
        stages.CompilableProgram,
        stages.ProgramSource[languages.NanobindSrcL, languages.LanguageWithHeaderFilesSettings],
    ],
):
    language_settings: Optional[languages.LanguageWithHeaderFilesSettings] = None
    # TODO replace by more general mechanism, see https://github.com/GridTools/gt4py/issues/1135
    enable_itir_transforms: bool = True
    use_imperative_backend: bool = False
    device_type: core_defs.DeviceType = core_defs.DeviceType.CPU
    symbolic_domain_sizes: Optional[dict[str, str]] = None

    def _default_language_settings(self) -> languages.LanguageWithHeaderFilesSettings:
        match self.device_type:
            case core_defs.DeviceType.CUDA:
                return languages.LanguageWithHeaderFilesSettings(
                    formatter_key=cpp_interface.CPP_DEFAULT.formatter_key,
                    formatter_style=cpp_interface.CPP_DEFAULT.formatter_style,
                    file_extension="cu",
                    header_extension="cuh",
                )
            case core_defs.DeviceType.ROCM:
                return languages.LanguageWithHeaderFilesSettings(
                    formatter_key=cpp_interface.CPP_DEFAULT.formatter_key,
                    formatter_style=cpp_interface.CPP_DEFAULT.formatter_style,
                    file_extension="hip",
                    header_extension="h",
                )
            case core_defs.DeviceType.CPU:
                return cpp_interface.CPP_DEFAULT
            case _:
                raise self._not_implemented_for_device_type()

    def _process_regular_arguments(
        self,
        program: itir.Program,
        arg_types: tuple[ts.TypeSpec, ...],
        offset_provider_type: common.OffsetProviderType,
    ) -> tuple[list[interface.Parameter], list[str]]:
        parameters: list[interface.Parameter] = []
        arg_exprs: list[str] = []

        for arg_type, program_param in zip(arg_types, program.params, strict=True):
            arg_type = arg_type.type_ if isinstance(arg_type, arguments.StaticArg) else arg_type
            # parameter
            parameter = get_param_description(program_param.id, arg_type)
            parameters.append(parameter)

            arg = f"std::forward<decltype({parameter.name})>({parameter.name})"

            if isinstance(parameter.type_, ts.FieldType):
                for dim in parameter.type_.dims:
                    if (
                        isinstance(
                            dim, fbuiltins.FieldOffset
                        )  # TODO(havogt): remove support for FieldOffset as Dimension
                        or dim.kind is common.DimensionKind.LOCAL
                    ):
                        # translate sparse dimensions to tuple dtype
                        dim_name = dim.value
                        connectivity = offset_provider_type[dim_name]
                        assert isinstance(connectivity, common.NeighborConnectivityType)
                        size = connectivity.max_neighbors
                        arg = f"gridtools::sid::dimension_to_tuple_like<generated::{dim_name}_t, {size}>({arg})"
            arg_exprs.append(arg)
        return parameters, arg_exprs

    def _process_connectivity_args(
        self, offset_provider_type: common.OffsetProviderType
    ) -> tuple[list[interface.Parameter], list[str]]:
        parameters: list[interface.Parameter] = []
        arg_exprs: list[str] = []

        for name, connectivity_type in offset_provider_type.items():
            if isinstance(connectivity_type, common.NeighborConnectivityType):
                if connectivity_type.dtype.scalar_type not in [np.int32, np.int64]:
                    raise ValueError(
                        "Neighbor table indices must be of type 'np.int32' or 'np.int64'."
                    )

                # parameter
                parameters.append(
                    interface.Parameter(
                        name=GENERATED_CONNECTIVITY_PARAM_PREFIX + name.lower(),
                        type_=ts.FieldType(
                            dims=list(connectivity_type.domain),
                            dtype=type_translation.from_dtype(connectivity_type.dtype),
                        ),
                    )
                )

                # connectivity argument expression
                nbtbl = (
                    f"gridtools::fn::sid_neighbor_table::as_neighbor_table<"
                    f"generated::{connectivity_type.domain[0].value}_t, "
                    f"generated::{connectivity_type.domain[1].value}_t, "
                    f"{connectivity_type.max_neighbors}"
                    f">(std::forward<decltype({GENERATED_CONNECTIVITY_PARAM_PREFIX}{name.lower()})>({GENERATED_CONNECTIVITY_PARAM_PREFIX}{name.lower()}))"
                )
                arg_exprs.append(
                    f"gridtools::hymap::keys<generated::{name}_t>::make_values({nbtbl})"
                )
            elif isinstance(connectivity_type, common.Dimension):
                pass
            else:
                raise AssertionError(
                    f"Expected offset provider type '{name}' to be a 'NeighborConnectivityType' or 'Dimension', "
                    f"got '{type(connectivity_type).__name__}'."
                )

        return parameters, arg_exprs

    def _preprocess_program(
        self,
        program: itir.Program,
        offset_provider: common.OffsetProvider | common.OffsetProviderType,
    ) -> itir.Program:
        apply_common_transforms = functools.partial(
            pass_manager.apply_common_transforms,
            extract_temporaries=True,
            offset_provider=offset_provider,
            # sid::composite (via hymap) supports assigning from tuple with more elements to tuple with fewer elements
            unconditionally_collapse_tuples=True,
            symbolic_domain_sizes=self.symbolic_domain_sizes,
        )

        new_program = apply_common_transforms(
            program, unroll_reduce=not self.use_imperative_backend
        )

        if self.use_imperative_backend and any(
            node.id == "neighbors"
            for node in new_program.pre_walk_values().if_isinstance(itir.SymRef)
        ):
            # if we don't unroll, there may be lifts left in the itir which can't be lowered to
            # gtfn. In this case, just retry with unrolled reductions.
            new_program = apply_common_transforms(program, unroll_reduce=True)

        return new_program

    def generate_stencil_source(
        self,
        program: itir.Program,
        offset_provider: common.OffsetProvider | common.OffsetProviderType,
        column_axis: Optional[common.Dimension],
    ) -> str:
        if self.enable_itir_transforms:
            new_program = self._preprocess_program(program, offset_provider)
        else:
            assert isinstance(program, itir.Program)
            new_program = program

        gtfn_ir = GTFN_lowering.apply(
            new_program,
            offset_provider_type=common.offset_provider_to_type(offset_provider),
            column_axis=column_axis,
        )

        if self.use_imperative_backend:
            gtfn_im_ir = GTFN_IM_lowering().visit(node=gtfn_ir)
            generated_code = GTFNIMCodegen.apply(gtfn_im_ir)
        else:
            generated_code = GTFNCodegen.apply(gtfn_ir)

        return codegen.format_source("cpp", generated_code, style="LLVM")

    def __call__(
        self, inp: stages.CompilableProgram
    ) -> stages.ProgramSource[languages.NanobindSrcL, languages.LanguageWithHeaderFilesSettings]:
        """Generate GTFN C++ code from the ITIR definition."""
        program: itir.Program = inp.data

        # handle regular parameters and arguments of the program (i.e. what the user defined in
        #  the program)
        arg_types = tuple(
            arg.type_ if isinstance(arg, arguments.StaticArg) else arg for arg in inp.args.args
        )
        regular_parameters, regular_args_expr = self._process_regular_arguments(
            program, arg_types, inp.args.offset_provider_type
        )

        # handle connectivity parameters and arguments (i.e. what the user provided in the offset
        #  provider)
        connectivity_parameters, connectivity_args_expr = self._process_connectivity_args(
            inp.args.offset_provider_type
        )

        # combine into a format that is aligned with what the backend expects
        parameters: list[interface.Parameter] = regular_parameters + connectivity_parameters
        backend_arg = self._backend_type()
        args_expr: list[str] = [backend_arg, *regular_args_expr]

        function = interface.Function(program.id, tuple(parameters))
        decl_body = (
            f"return generated::{function.name}("
            f"{', '.join(connectivity_args_expr)})({', '.join(args_expr)});"
        )
        decl_src = cpp_interface.render_function_declaration(function, body=decl_body)
        stencil_src = self.generate_stencil_source(
            program,
            inp.args.offset_provider,
            inp.args.column_axis,
        )
        source_code = interface.format_source(
            self._language_settings(),
            f"""
                    #include <{self._backend_header()}>
                    #include <gridtools/sid/dimension_to_tuple_like.hpp>
                    {stencil_src}
                    {decl_src}
                    """.strip(),
        )

        module: stages.ProgramSource[
            languages.NanobindSrcL, languages.LanguageWithHeaderFilesSettings
        ] = stages.ProgramSource(
            entry_point=function,
            library_deps=(interface.LibraryDependency(self._library_name(), "master"),),
            source_code=source_code,
            language=self._language(),
            language_settings=self._language_settings(),
            implicit_domain=inp.data.implicit_domain,
        )
        return module

    def _backend_header(self) -> str:
        match self.device_type:
            case core_defs.DeviceType.CUDA | core_defs.DeviceType.ROCM:
                return "gridtools/fn/backend/gpu.hpp"
            case core_defs.DeviceType.CPU:
                return "gridtools/fn/backend/naive.hpp"
            case _:
                raise self._not_implemented_for_device_type()

    def _backend_type(self) -> str:
        match self.device_type:
            case core_defs.DeviceType.CUDA | core_defs.DeviceType.ROCM:
                return "gridtools::fn::backend::gpu<generated::block_sizes_t>{}"
            case core_defs.DeviceType.CPU:
                return "gridtools::fn::backend::naive{}"
            case _:
                raise self._not_implemented_for_device_type()

    def _language(self) -> type[languages.NanobindSrcL]:
        match self.device_type:
            case core_defs.DeviceType.CUDA:
                return languages.CUDA
            case core_defs.DeviceType.ROCM:
                return languages.HIP
            case core_defs.DeviceType.CPU:
                return languages.CPP
            case _:
                raise self._not_implemented_for_device_type()

    def _language_settings(self) -> languages.LanguageWithHeaderFilesSettings:
        return (
            self.language_settings
            if self.language_settings is not None
            else self._default_language_settings()
        )

    def _library_name(self) -> str:
        match self.device_type:
            case core_defs.DeviceType.CUDA | core_defs.DeviceType.ROCM:
                return "gridtools_gpu"
            case core_defs.DeviceType.CPU:
                return "gridtools_cpu"
            case _:
                raise self._not_implemented_for_device_type()

    def _not_implemented_for_device_type(self) -> NotImplementedError:
        return NotImplementedError(
            f"{self.__class__.__name__} is not implemented for device type {self.device_type.name}"
        )


class GTFNTranslationStepFactory(factory.Factory):
    class Meta:
        model = GTFNTranslationStep


translate_program_cpu: Final[step_types.TranslationStep] = GTFNTranslationStepFactory()

translate_program_gpu: Final[step_types.TranslationStep] = GTFNTranslationStepFactory(
    device_type=core_defs.DeviceType.CUDA
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from .simple_assign import SingleAssignTargetPass
from .single_static_assign import SingleStaticAssignPass
from .stringify_annotations import StringifyAnnotationsPass
from .unchain_compares import UnchainComparesPass


__all__ = [
    "SingleAssignTargetPass",
    "SingleStaticAssignPass",
    "StringifyAnnotationsPass",
    "UnchainComparesPass",
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from warnings import warn

from .base import REGISTRY, Backend, BaseBackend, BasePyExtBackend, from_name, register
from .cuda_backend import CudaBackend
from .debug_backend import DebugBackend
from .gtcpp_backend import GTCpuIfirstBackend, GTCpuKfirstBackend, GTGpuBackend
from .module_generator import BaseModuleGenerator
from .numpy_backend import NumpyBackend


__all__ = [
    "REGISTRY",
    "Backend",
    "BaseBackend",
    "BaseModuleGenerator",
    "BasePyExtBackend",
    "CudaBackend",
    "DebugBackend",
    "GTCpuIfirstBackend",
    "GTCpuKfirstBackend",
    "GTGpuBackend",
    "NumpyBackend",
    "from_name",
    "register",
]


try:
    from .dace_backend import DaceCPUBackend, DaceGPUBackend

    __all__ += ["DaceCPUBackend", "DaceGPUBackend"]
except ImportError:
    warn(
        "GT4Py was unable to load DaCe. DaCe backends (`dace:cpu` and `dace:gpu`) will not be available.",
        stacklevel=2,
    )



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, TypeAlias

import dace
from dace import data as dace_data
from dace.sdfg import nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


PropagatedStrideRecord: TypeAlias = tuple[str, dace_nodes.NestedSDFG]
"""Record of a stride that has been propagated into a NestedSDFG.

The type combines the NestedSDFG into which the strides were already propagated
and the data within that NestedSDFG to which we have propagated the strides,
which is the connector name on the NestedSDFG.
We need the NestedSDFG because we have to know what was already processed,
however, we also need the inner array name because of aliasing, i.e. a data
descriptor on the outside could be mapped to multiple data descriptors
inside the NestedSDFG.
"""


def gt_change_transient_strides(
    sdfg: dace.SDFG,
    gpu: bool,
) -> dace.SDFG:
    """Modifies the strides of transients.

    The function will analyse the access patterns and set the strides of
    transients in the optimal way.
    The function should run after all maps have been created.

    After the strides have been adjusted the function will also propagate
    the strides into nested SDFG. This propagation will happen with
    `ignore_symbol_mapping` set to `True`, see `gt_propagate_strides_of()`
    for more.

    Args:
        sdfg: The SDFG to process.
        gpu: If the SDFG is supposed to run on the GPU.

    Note:
        Currently the function will not scan the access pattern. Instead it will
        either use FORTRAN order for GPU or C order (which is assumed to be the
        default, so it is a no ops).

    Todo:
        - Implement the estimation correctly.
    """
    # TODO(phimeull): Implement this function correctly.

    # We assume that by default we have C order which is already correct,
    #  so in this case we have a no ops
    if not gpu:
        return sdfg

    for nsdfg in sdfg.all_sdfgs_recursive():
        _gt_change_transient_strides_non_recursive_impl(nsdfg)


def _gt_change_transient_strides_non_recursive_impl(
    sdfg: dace.SDFG,
) -> None:
    """Set optimal strides of all transients in the SDFG.

    The function will look for all top level transients, see `_gt_find_toplevel_data_accesses()`
    and set their strides such that the access is optimal, see Note. The function
    will also run `gt_propagate_strides_of()` to propagate the strides into nested SDFGs.

    This function should never be called directly but always through
    `gt_change_transient_strides()`!

    Note:
        Currently the function just reverses the strides of the data descriptor
        it processes. Since DaCe generates `C` order by default this lead to
        FORTRAN order, which is (for now) sufficient to optimize the memory
        layout to GPU.

    Todo:
        Make this function more intelligent to analyse the access pattern and then
        figuring out the best order.
    """
    # NOTE: Processing the transient here is enough. If we are inside a
    #   NestedSDFG then they were handled before on the level above us.
    top_level_transients_and_their_accesses = _gt_find_toplevel_data_accesses(
        sdfg=sdfg,
        only_transients=True,
        only_arrays=True,
    )
    for top_level_transient, accesses in top_level_transients_and_their_accesses.items():
        desc: dace_data.Array = sdfg.arrays[top_level_transient]

        # Setting the strides only make sense if we have more than one dimensions
        ndim = len(desc.shape)
        if ndim <= 1:
            continue

        # We assume that everything is in C order initially, to get FORTRAN order
        #  we simply have to reverse the order.
        # TODO(phimuell): Improve this.
        new_stride_order = list(range(ndim))
        desc.set_strides_from_layout(*new_stride_order)

        # Now we have to propagate the changed strides. Because we already have
        #  collected all the AccessNodes we are using the
        #  `gt_propagate_strides_from_access_node()` function, but we have to
        #  create `processed_nsdfg` set already outside here.
        #  Furthermore, the same comment as above applies here, we do not have to
        #  propagate the non-transients, because they either come from outside,
        #  or they were already handled in the levels above, where they were
        #  defined and then propagated down.
        # TODO(phimuell): Updated the functions such that only one scan is needed.
        processed_nsdfgs: set[dace_nodes.NestedSDFG] = set()
        for state, access_node in accesses:
            gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=state,
                outer_node=access_node,
                processed_nsdfgs=processed_nsdfgs,
                ignore_symbol_mapping=True,
            )


def gt_propagate_strides_of(
    sdfg: dace.SDFG,
    data_name: str,
    ignore_symbol_mapping: bool = True,
) -> None:
    """Propagates the strides of `data_name` within the whole SDFG.

    This function will call `gt_propagate_strides_from_access_node()` for every
    AccessNode that refers to `data_name`. It will also make sure that a descriptor
    inside a NestedSDFG is only processed once.

    Args:
        sdfg: The SDFG on which we operate.
        data_name: Name of the data descriptor that should be handled.
        ignore_symbol_mapping: If `False` (default is `True`) try to modify the `symbol_mapping`
            of NestedSDFGs instead of manipulating the data descriptor.
    """

    # Defining it here ensures that we will not enter an NestedSDFG multiple times.
    processed_nsdfgs: set[PropagatedStrideRecord] = set()

    for state in sdfg.states():
        for dnode in state.data_nodes():
            if dnode.data != data_name:
                continue
            gt_propagate_strides_from_access_node(
                sdfg=sdfg,
                state=state,
                outer_node=dnode,
                processed_nsdfgs=processed_nsdfgs,
                ignore_symbol_mapping=ignore_symbol_mapping,
            )


def gt_propagate_strides_from_access_node(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    outer_node: dace_nodes.AccessNode,
    ignore_symbol_mapping: bool = True,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]] = None,
) -> None:
    """Propagates the stride of `outer_node` to any adjacent NestedSDFG.

    The function will propagate the strides of the data descriptor `outer_node`
    refers to along all adjacent edges of `outer_node`. If one of these edges
    leads to a NestedSDFG then the function will modify the strides of data
    descriptor within to match the strides on the outside. The function will then
    recursively process NestedSDFG.

    It is important that this function will only handle the NestedSDFGs that are
    reachable from `outer_node`. To fully propagate the strides the
    `gt_propagate_strides_of()` should be used.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that reads from the data node, the nested SDFG is expected as the destination.
        outer_node: The data node whose strides should be propagated.
        ignore_symbol_mapping: If `False` (default is `True`), try to modify the `symbol_mapping`
            of NestedSDFGs instead of manipulating the data descriptor.
        processed_nsdfgs: Set of NestedSDFG that were already processed and will be ignored.
            Only specify when you know what your are doing.
    """
    assert isinstance(state, dace.SDFGState)

    if processed_nsdfgs is None:
        # For preventing the case that nested SDFGs are handled multiple time.
        processed_nsdfgs = set()

    for in_edge in state.in_edges(outer_node):
        gt_map_strides_to_src_nested_sdfg(
            sdfg=sdfg,
            state=state,
            edge=in_edge,
            outer_node=outer_node,
            processed_nsdfgs=processed_nsdfgs,
            ignore_symbol_mapping=ignore_symbol_mapping,
        )
    for out_edge in state.out_edges(outer_node):
        gt_map_strides_to_dst_nested_sdfg(
            sdfg=sdfg,
            state=state,
            edge=out_edge,
            outer_node=outer_node,
            processed_nsdfgs=processed_nsdfgs,
            ignore_symbol_mapping=ignore_symbol_mapping,
        )


def gt_map_strides_to_dst_nested_sdfg(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace.sdfg.graph.Edge,
    outer_node: dace.nodes.AccessNode,
    ignore_symbol_mapping: bool = True,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]] = None,
) -> None:
    """Propagates the strides of `outer_node` along `edge` in the dataflow direction.

    In this context "along the dataflow direction" means that `edge` is an outgoing
    edge of `outer_node` and the strides are propagated into all NestedSDFGs that
    are downstream of `outer_node`.

    Except in certain cases this function should not be used directly. It is
    instead recommended to use `gt_propagate_strides_of()`, which propagates
    all edges in the SDFG.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that writes to the data node, the nested SDFG is expected as the source.
        outer_node: The data node whose strides should be propagated.
        ignore_symbol_mapping: If `False`, the default, try to modify the `symbol_mapping`
            of NestedSDFGs instead of manipulating the data descriptor.
        processed_nsdfgs: Set of NestedSDFGs that were already processed. Only specify when
            you know what your are doing.
    """
    assert edge.src is outer_node
    _gt_map_strides_to_nested_sdfg_src_dst(
        sdfg=sdfg,
        state=state,
        edge=edge,
        outer_node=outer_node,
        processed_nsdfgs=processed_nsdfgs,
        propagate_along_dataflow=True,
        ignore_symbol_mapping=ignore_symbol_mapping,
    )


def gt_map_strides_to_src_nested_sdfg(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace.sdfg.graph.Edge,
    outer_node: dace.nodes.AccessNode,
    ignore_symbol_mapping: bool = False,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]] = None,
) -> None:
    """Propagates the strides of `outer_node` along `edge` in the opposite direction of the dataflow

    In this context "in the opposite direction of the dataflow" means that `edge`
    is an incoming edge of `outer_node` and the strides are propagated into all
    NestedSDFGs that are upstream of `outer_node`.

    Except in certain cases this function should not be used directly. It is
    instead recommended to use `gt_propagate_strides_of()`, which propagates
    all edges in the SDFG.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that writes to the data node, the nested SDFG is expected as the source.
        outer_node: The data node whose strides should be propagated.
        ignore_symbol_mapping: If `False`, the default, try to modify the `symbol_mapping`
            of NestedSDFGs instead of manipulating the data descriptor.
        processed_nsdfgs: Set of NestedSDFGs that were already processed. Only specify when
            you know what your are doing.
    """
    _gt_map_strides_to_nested_sdfg_src_dst(
        sdfg=sdfg,
        state=state,
        edge=edge,
        outer_node=outer_node,
        processed_nsdfgs=processed_nsdfgs,
        propagate_along_dataflow=False,
        ignore_symbol_mapping=ignore_symbol_mapping,
    )


def _gt_map_strides_to_nested_sdfg_src_dst(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
    outer_node: dace.nodes.AccessNode,
    processed_nsdfgs: Optional[set[PropagatedStrideRecord]],
    propagate_along_dataflow: bool,
    ignore_symbol_mapping: bool = False,
) -> None:
    """Propagates the stride of `outer_node` along `edge`.

    The function will follow `edge`, the direction depends on the value of
    `propagate_along_dataflow` and propagate the strides of `outer_node`
    into every NestedSDFG that is reachable by following `edge`.

    When the function encounters a NestedSDFG it will determine what data
    the `outer_node` is mapped to on the inside of the NestedSDFG.
    It will then replace the stride of the inner descriptor with the ones
    of the outside. Afterwards it will recursively propagate the strides
    inside the NestedSDFG.
    During this propagation the function will follow any edges.

    If the function reaches a NestedSDFG that is listed inside `processed_nsdfgs`
    then it will be skipped. NestedSDFGs that have been processed will be added
    to the `processed_nsdfgs`.

    Args:
        sdfg: The SDFG to process.
        state: The state where the data node is used.
        edge: The edge that reads from the data node, the nested SDFG is expected as the destination.
        outer_node: The data node whose strides should be propagated.
        processed_nsdfgs: Set of Nested SDFG that were already processed and will be ignored.
            Only specify when you know what your are doing.
        propagate_along_dataflow: Determine the direction of propagation. If `True` the
            function follows the dataflow.
        ignore_symbol_mapping: If `False`, the default, try to modify the `symbol_mapping`
            of NestedSDFGs instead of manipulating the data descriptor.

    Note:
        A user should not use this function directly, instead `gt_propagate_strides_of()`,
        `gt_map_strides_to_src_nested_sdfg()` (`propagate_along_dataflow == `False`)
        or `gt_map_strides_to_dst_nested_sdfg()` (`propagate_along_dataflow == `True`)
        should be used.

    Todo:
        Try using `MemletTree` for the propagation.
    """
    # If `processed_nsdfg` is `None` then this is the first call. We will now
    #  allocate the `set` and pass it as argument to all recursive calls, this
    #  ensures that the `set` is the same everywhere.
    if processed_nsdfgs is None:
        processed_nsdfgs = set()

    if propagate_along_dataflow:
        # Propagate along the dataflow or forward, so we are interested at the `dst` of the edge.
        ScopeNode = dace_nodes.MapEntry

        def get_node(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> dace_nodes.Node:
            return edge.dst

        def get_inner_data(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> str:
            return edge.dst_conn

        def get_subset(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> dace.subsets.Subset:
            return edge.data.get_src_subset(edge, state)

        def next_edges_by_connector(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> list[dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]]:
            if edge.dst_conn is None or not edge.dst_conn.startswith("IN_"):
                return []
            return list(state.out_edges_by_connector(edge.dst, "OUT_" + edge.dst_conn[3:]))

    else:
        # Propagate against the dataflow or backward, so we are interested at the `src` of the edge.
        ScopeNode = dace_nodes.MapExit

        def get_node(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> dace_nodes.Node:
            return edge.src

        def get_inner_data(edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]) -> str:
            return edge.src_conn

        def get_subset(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> dace.subsets.Subset:
            return edge.data.get_dst_subset(edge, state)

        def next_edges_by_connector(
            state: dace.SDFGState,
            edge: dace.sdfg.graph.MultiConnectorEdge[dace.Memlet],
        ) -> list[dace.sdfg.graph.MultiConnectorEdge[dace.Memlet]]:
            return list(state.in_edges_by_connector(edge.src, "IN_" + edge.src_conn[4:]))

    if isinstance(get_node(edge), ScopeNode):
        for next_edge in next_edges_by_connector(state, edge):
            _gt_map_strides_to_nested_sdfg_src_dst(
                sdfg=sdfg,
                state=state,
                edge=next_edge,
                outer_node=outer_node,
                processed_nsdfgs=processed_nsdfgs,
                propagate_along_dataflow=propagate_along_dataflow,
                ignore_symbol_mapping=ignore_symbol_mapping,
            )

    elif isinstance(get_node(edge), dace.nodes.NestedSDFG):
        nsdfg_node = get_node(edge)
        inner_data = get_inner_data(edge)
        process_record = (inner_data, nsdfg_node)

        if process_record in processed_nsdfgs:
            # We already handled this NestedSDFG and the inner data.
            return

        # Mark this nested SDFG as processed.
        processed_nsdfgs.add(process_record)

        # Now set the stride of the data descriptor inside the nested SDFG to
        #  the ones it has outside.
        _gt_map_strides_into_nested_sdfg(
            sdfg=sdfg,
            nsdfg_node=nsdfg_node,
            inner_data=inner_data,
            outer_subset=get_subset(state, edge),
            outer_desc=outer_node.desc(sdfg),
            ignore_symbol_mapping=ignore_symbol_mapping,
        )

        # Since the function call above is not recursive we have now to propagate
        #  the change into the NestedSDFGs. Using `_gt_find_toplevel_data_accesses()`
        #  is a bit overkill, but allows for a more uniform processing.
        # TODO(phimuell): Instead of scanning every level for every data we modify
        #   we should scan the whole SDFG once and then reuse this information.
        accesses_in_nested_sdfg = _gt_find_toplevel_data_accesses(
            sdfg=nsdfg_node.sdfg,
            only_transients=False,  # Because on the nested levels they are globals.
            only_arrays=True,
        )
        for nested_state, nested_access in accesses_in_nested_sdfg.get(inner_data, list()):
            # We have to use `gt_propagate_strides_from_access_node()` here because we
            #  have to handle its entirety. We could wait until the other branch processes
            #  the nested SDFG, but this might not work, so let's do it fully now.
            gt_propagate_strides_from_access_node(
                sdfg=nsdfg_node.sdfg,
                state=nested_state,
                outer_node=nested_access,
                processed_nsdfgs=processed_nsdfgs,
                ignore_symbol_mapping=ignore_symbol_mapping,
            )


def _gt_map_strides_into_nested_sdfg(
    sdfg: dace.SDFG,
    nsdfg_node: dace.nodes.NestedSDFG,
    inner_data: str,
    outer_subset: dace.subsets.Subset,
    outer_desc: dace_data.Data,
    ignore_symbol_mapping: bool,
) -> None:
    """Modify the strides of `inner_data` inside `nsdfg_node` to match `outer_desc`.

    `inner_data` is the name of a data descriptor inside the NestedSDFG.
    The function will then modify the strides of `inner_data`, assuming this
    is an array, to match the ones of `outer_desc`.

    Args:
        sdfg: The SDFG containing the NestedSDFG.
        nsdfg_node: The node in the parent SDFG that contains the NestedSDFG.
        inner_data: The name of the data descriptor that should be processed
            inside the NestedSDFG (by construction also a connector name).
        outer_subset: The subset that describes what part of the outer data is
            mapped into the NestedSDFG.
        outer_desc: The data descriptor of the data on the outside.
        ignore_symbol_mapping: If possible the function will perform the renaming
            through the `symbol_mapping` of the nested SDFG. If `True` then
            the function will always perform the renaming.
            Note that setting this value to `False` might have negative side effects.

    Todo:
        - Handle explicit dimensions of size 1.
        - What should we do if the stride symbol is used somewhere else, creating an
            alias is probably not the right thing?
        - Handle the case if the outer stride symbol is already used in another
            context inside the Neste SDFG.
    """
    # We need to compute the new strides. In the following we assume that the
    #  relative order of the dimensions does not change, but we support the case
    #  where some dimensions of the outer data descriptor are not present on the
    #  inside. For example this happens for the Memlet `a[__i0, 0:__a_size1]`. We
    #  detect this case by checking if the Memlet subset in that dimension has size 1.
    # TODO(phimuell): Handle the case were some additional size 1 dimensions are added.
    inner_desc: dace_data.Data = nsdfg_node.sdfg.arrays[inner_data]
    inner_shape = inner_desc.shape
    inner_strides_init = inner_desc.strides

    outer_shape = outer_desc.shape
    outer_strides = outer_desc.strides
    outer_inflow = outer_subset.size()

    if isinstance(inner_desc, dace_data.Scalar):
        # A scalar does not have a stride that must be propagated.
        return

    # Now determine the new stride that is needed on the inside.
    new_strides: list = []
    if len(outer_shape) == len(inner_shape):
        # The inner and the outer descriptor have the same dimensionality.
        #  We now have to decide if we should take the stride from the outside,
        #  which happens for example in case of `A[0:N, 0:M] -> B[N, M]`, or if we
        #  must take 1, which happens if we do `A[0:N, i] -> B[N, 1]`, we detect that
        #  based on the volume that flows in.
        for dim_ostride, dim_oinflow in zip(outer_strides, outer_inflow, strict=True):
            new_strides.append(1 if dim_oinflow == 1 else dim_ostride)

    elif len(inner_shape) < len(outer_shape):
        # There are less dimensions on the inside than on the outside. This means
        #  that some were sliced away. We detect this case by checking if the Memlet
        #  subset in that dimension has size 1.
        #  NOTE: That this is not always correct as it might be possible that there
        #   are some explicit size 1 dimensions at several places.
        new_strides = []
        for dim_ostride, dim_oinflow in zip(outer_strides, outer_inflow, strict=True):
            if dim_oinflow == 1:
                pass
            else:
                new_strides.append(dim_ostride)
            assert len(new_strides) <= len(inner_shape)
    else:
        # The case that we have more dimensions on the inside than on the outside.
        #  This is currently not supported.
        raise NotImplementedError("NestedSDFGs can not be used to increase the rank.")

    if len(new_strides) != len(inner_shape):
        raise ValueError("Failed to compute the inner strides.")

    # Now we actually replace the strides, there are two ways of doing it.
    #  The first is to create an alias in the `symbol_mapping`, however,
    #  this is only possible if the current strides are singular symbols,
    #  like `__a_strides_1`, but not expressions such as `horizontal_end - horizontal_start`
    #  or literal values. Furthermore, this would change the meaning of the
    #  old stride symbol in any context and not only in the one of the stride
    #  of a single and isolated data descriptor.
    #  The second way would be to replace `strides` attribute of the
    #  inner data descriptor. In case the new stride consists of expressions
    #  such as `value1 - value2` we have to make them available inside the
    #  NestedSDFG. However, it could be that the strides is used somewhere else.
    # We will do the following, if `ignore_symbol_mapping` is `False` and
    #  the strides of the inner descriptors are symbols, we will use the
    #  symbol mapping. Otherwise, we will replace the `strides` attribute
    #  of the inner descriptor, in addition we will install a remapping,
    #  for those values that were a symbol.
    if (not ignore_symbol_mapping) and all(
        isinstance(inner_stride, dace.symbol) for inner_stride in inner_strides_init
    ):
        # Use the symbol
        for inner_stride, outer_stride in zip(inner_desc.strides, new_strides, strict=True):
            nsdfg_node.symbol_mapping[inner_stride.name] = outer_stride
    else:
        # We have to replace the `strides` attribute of the inner descriptor.
        inner_desc.set_shape(inner_desc.shape, new_strides)

        # Now find the free symbols that the new strides need.
        #  Note that usually `free_symbols` returns `set[str]`, but here, because
        #  we fall back on SymPy, we get back symbols. We will keep them, because
        #  then we can use them to extract the type form them, which we need later.
        new_strides_symbols: list[dace.symbol] = []
        for new_stride_dim in new_strides:
            if dace.symbolic.issymbolic(new_stride_dim):
                new_strides_symbols.extend(sym for sym in new_stride_dim.free_symbols)
            else:
                # It is not already a symbol, so we turn it into a symbol.
                #  However, we only add it, if it is also a symbol, for example `1`.
                #  should not be added.
                new_stride_symbol = dace.symbolic.pystr_to_symbolic(new_stride_dim)
                if new_stride_symbol.is_symbol:
                    new_strides_symbols.append(new_stride_symbol)

        # Now we determine the set of symbols that should be mapped inside the NestedSDFG.
        #  We will exclude all that are already inside the `symbol_mapping` (we do not
        #  check if they map to the same value, we just hope it). Furthermore,
        #  we will exclude all symbols that are listed in the `symbols` property
        #  of the SDFG that is nested, and hope that it has the same meaning.
        # TODO(phimuell): Add better checks to avoid overwriting.
        missing_symbol_mappings: set[dace.symbol] = {
            sym
            for sym in new_strides_symbols
            if not (sym.name in nsdfg_node.sdfg.symbols or sym.name in nsdfg_node.symbol_mapping)
        }

        # Now propagate the symbols from the parent SDFG to the NestedSDFG.
        for sym in missing_symbol_mappings:
            assert sym.name in sdfg.symbols, f"Expected that '{sym}' is defined in the parent SDFG."
            nsdfg_node.sdfg.add_symbol(sym.name, sdfg.symbols[sym.name])
            nsdfg_node.symbol_mapping[sym.name] = sym


def _gt_find_toplevel_data_accesses(
    sdfg: dace.SDFG,
    only_transients: bool,
    only_arrays: bool = False,
) -> dict[str, list[tuple[dace.SDFGState, dace_nodes.AccessNode]]]:
    """Find all data that is accessed on the top level.

    The function will scan the SDFG, ignoring nested one, and return the
    name of all data that only have AccessNodes on the top level. In data
    is found that has an AccessNode on both the top level and in a nested
    scope and error is generated.
    By default the function will return transient and non transient data,
    however, if `only_transients` is `True` then only transient data will
    be returned.
    Furthermore, the function will ignore an access in the following cases:
    - The AccessNode refers to data that is a register.
    - The AccessNode refers to a View.

    Args:
        sdfg: The SDFG to process.
        only_transients: If `True` only include transients.
        only_arrays: If `True`, defaults to `False`, only arrays are returned.

    Returns:
        A `dict` that maps the name of a data container, to a list of tuples
        containing the state where the AccessNode was found and the AccessNode.
    """
    # List of data that is accessed on the top level and all its access node.
    top_level_data: dict[str, list[tuple[dace.SDFGState, dace_nodes.AccessNode]]] = dict()

    # List of all data that were found not on top level.
    not_top_level_data: set[str] = set()

    for state in sdfg.states():
        assert isinstance(state, dace.SDFGState)
        scope_dict = state.scope_dict()
        for dnode in state.data_nodes():
            data: str = dnode.data
            if scope_dict[dnode] is not None:
                # The node was not found on the top level. So we can ignore it.
                # We also check if it was ever found on the top level, this should
                #  not happen, as everything should go through Maps. But some strange
                #  DaCe transformation might do it.
                assert data not in top_level_data, (
                    f"Found {data} on the top level and inside a scope."
                )
                not_top_level_data.add(data)
                continue

            elif data in top_level_data:
                # The data is already known to be in top level data, so we must add the
                #  AccessNode to the list of known nodes. But nothing else.
                top_level_data[data].append((state, dnode))
                continue

            elif gtx_transformations.utils.is_view(dnode, sdfg):
                # The AccessNode refers to a View so we ignore it anyway.
                continue

            # We have found a new data node that is on the top node and is unknown.
            assert data not in not_top_level_data, (
                f"Found {data} on the top level and inside a scope."
            )
            desc: dace_data.Data = dnode.desc(sdfg)

            # Check if we only accept arrays
            if only_arrays and not isinstance(desc, dace_data.Array):
                continue

            # For now we ignore registers.
            #  We do this because register are allocated on the stack, so the compiler
            #  has all information and should organize the best thing possible.
            # TODO(phimuell): verify this.
            elif desc.storage is dace.StorageType.Register:
                continue

            # We are only interested in transients
            if only_transients and (not desc.transient):
                continue

            # Now create the new entry in the list and record the AccessNode.
            top_level_data[data] = [(state, dnode)]
    return top_level_data



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, List, TypeAlias

from dace import data, dtypes, symbolic

from gt4py import eve
from gt4py.cartesian.gtc import common, definitions, oir
from gt4py.cartesian.gtc.dace import oir_to_tasklet, treeir as tir, utils
from gt4py.cartesian.gtc.passes.gtir_k_boundary import compute_k_boundary
from gt4py.cartesian.gtc.passes.oir_optimizations import utils as oir_utils
from gt4py.cartesian.stencil_builder import StencilBuilder


ControlFlow: TypeAlias = (
    oir.HorizontalExecution | oir.While | oir.MaskStmt | oir.HorizontalRestriction
)
"""All control flow OIR nodes"""

DEFAULT_STORAGE_TYPE = {
    dtypes.DeviceType.CPU: dtypes.StorageType.Default,
    dtypes.DeviceType.GPU: dtypes.StorageType.GPU_Global,
}
"""Default dace residency types per device type."""

DEFAULT_MAP_SCHEDULE = {
    dtypes.DeviceType.CPU: dtypes.ScheduleType.Default,
    dtypes.DeviceType.GPU: dtypes.ScheduleType.GPU_Device,
}
"""Default kernel target per device type."""


class OIRToTreeIR(eve.NodeVisitor):
    """
    Translate the GT4Py OIR into a Dace-centric TreeIR.

    TreeIR is built to be a minimum representation of DaCe's Schedule
    Tree. No transformation is done on TreeIR, though should be done
    once the TreeIR has been properly turned into a Schedule Tree.

    This class _does not_ deal with Tasklet representation, it defers that
    work to the OIRToTasklet visitor.
    """

    def __init__(self, builder: StencilBuilder) -> None:
        device_type_translate = {
            "CPU": dtypes.DeviceType.CPU,
            "GPU": dtypes.DeviceType.GPU,
        }

        device_type = builder.backend.storage_info["device"]
        if device_type.upper() not in device_type_translate:
            raise ValueError(f"Unknown device type {device_type}.")

        self._device_type = device_type_translate[device_type.upper()]
        self._api_signature = builder.gtir.api_signature
        self._k_bounds = compute_k_boundary(builder.gtir)

    def visit_CodeBlock(self, node: oir.CodeBlock, ctx: tir.Context) -> None:
        dace_tasklet, inputs, outputs = oir_to_tasklet.OIRToTasklet().visit_CodeBlock(
            node, root=ctx.root
        )

        tasklet = tir.Tasklet(
            tasklet=dace_tasklet,
            inputs=inputs,
            outputs=outputs,
            parent=ctx.current_scope,
        )
        ctx.current_scope.children.append(tasklet)

    def _group_statements(self, node: ControlFlow) -> list[oir.CodeBlock | ControlFlow]:
        """
        Group the body of a control flow node into CodeBlocks and other ControlFlow.

        This function only groups statements. The job of visiting the groups statements is
        left to the caller.
        """
        statements: List[ControlFlow | oir.CodeBlock | common.Stmt] = []
        groups: List[ControlFlow | oir.CodeBlock] = []

        for statement in node.body:
            if isinstance(statement, ControlFlow):
                if statements != []:
                    groups.append(
                        oir.CodeBlock(label=f"he_{id(node)}_{len(groups)}", body=statements)
                    )
                groups.append(statement)
                statements = []
            else:
                statements.append(statement)

        if statements != []:
            groups.append(oir.CodeBlock(label=f"he_{id(node)}_{len(groups)}", body=statements))

        return groups

    def _insert_evaluation_tasklet(
        self, node: oir.MaskStmt | oir.While, ctx: tir.Context
    ) -> tuple[str, oir.AssignStmt]:
        """Evaluate condition in a separate tasklet to avoid sympy problems down the line."""

        prefix = "while" if isinstance(node, oir.While) else "if"
        condition_name = f"{prefix}_condition_{id(node)}"

        ctx.root.containers[condition_name] = data.Scalar(
            utils.data_type_to_dace_typeclass(common.DataType.BOOL),
            transient=True,
            storage=dtypes.StorageType.Register,
            debuginfo=utils.get_dace_debuginfo(node),
        )

        assignment = oir.AssignStmt(
            left=oir.ScalarAccess(name=condition_name),
            right=node.cond if isinstance(node, oir.While) else node.mask,
        )

        code_block = oir.CodeBlock(label=f"masklet_{id(node)}", body=[assignment])
        self.visit(code_block, ctx=ctx)

        return (condition_name, assignment)

    def visit_HorizontalExecution(self, node: oir.HorizontalExecution, ctx: tir.Context) -> None:
        block_extent = ctx.block_extents[id(node)]

        axis_start_i = f"{block_extent[0][0]}"
        axis_start_j = f"{block_extent[1][0]}"
        axis_end_i = f"({tir.Axis.I.domain_dace_symbol()})  + ({block_extent[0][1]})"
        axis_end_j = f"({tir.Axis.J.domain_dace_symbol()})  + ({block_extent[1][1]})"

        loop = tir.HorizontalLoop(
            bounds_i=tir.Bounds(start=axis_start_i, end=axis_end_i),
            bounds_j=tir.Bounds(start=axis_start_j, end=axis_end_j),
            schedule=DEFAULT_MAP_SCHEDULE[self._device_type],
            children=[],
            parent=ctx.current_scope,
        )

        with loop.scope(ctx):
            # Push local scalars to the tree repository
            for local_scalar in node.declarations:
                ctx.root.containers[local_scalar.name] = data.Scalar(
                    dtype=utils.data_type_to_dace_typeclass(local_scalar.dtype),
                    transient=True,
                    storage=dtypes.StorageType.Register,
                    debuginfo=utils.get_dace_debuginfo(local_scalar),
                )

            groups = self._group_statements(node)
            self.visit(groups, ctx=ctx)

    def visit_MaskStmt(self, node: oir.MaskStmt, ctx: tir.Context) -> None:
        condition_name, _ = self._insert_evaluation_tasklet(node, ctx)

        if_else = tir.IfElse(
            if_condition_code=condition_name, children=[], parent=ctx.current_scope
        )

        with if_else.scope(ctx):
            groups = self._group_statements(node)
            self.visit(groups, ctx=ctx)

    def visit_HorizontalRestriction(
        self, node: oir.HorizontalRestriction, ctx: tir.Context
    ) -> None:
        """Translate `region` concept into If control flow in TreeIR."""
        condition_code = self.visit(node.mask, ctx=ctx)
        if_else = tir.IfElse(
            if_condition_code=condition_code, children=[], parent=ctx.current_scope
        )

        with if_else.scope(ctx):
            groups = self._group_statements(node)
            self.visit(groups, ctx=ctx)

    def visit_HorizontalMask(self, node: common.HorizontalMask, ctx: tir.Context) -> str:
        loop_i = tir.Axis.I.iteration_symbol()
        loop_j = tir.Axis.J.iteration_symbol()

        axis_start_i = "0"
        axis_end_i = tir.Axis.I.domain_symbol()
        axis_start_j = "0"
        axis_end_j = tir.Axis.J.domain_symbol()

        conditions: list[str] = []
        if node.i.start is not None:
            conditions.append(
                f"{loop_i} >= {self.visit(node.i.start, axis_start=axis_start_i, axis_end=axis_end_i)}"
            )
        if node.i.end is not None:
            conditions.append(
                f"{loop_i} < {self.visit(node.i.end, axis_start=axis_start_i, axis_end=axis_end_i)}"
            )
        if node.j.start is not None:
            conditions.append(
                f"{loop_j} >= {self.visit(node.j.start, axis_start=axis_start_j, axis_end=axis_end_j)}"
            )
        if node.j.end is not None:
            conditions.append(
                f"{loop_j} < {self.visit(node.j.end, axis_start=axis_start_j, axis_end=axis_end_j)}"
            )

        return " and ".join(conditions)

    def visit_While(self, node: oir.While, ctx: tir.Context) -> None:
        condition_name, assignment = self._insert_evaluation_tasklet(node, ctx)

        # Re-evaluate the condition as last step of the while loop
        node.body.append(assignment)

        # Use the mask created for conditional check
        while_ = tir.While(
            condition_code=condition_name,
            children=[],
            parent=ctx.current_scope,
        )

        with while_.scope(ctx):
            groups = self._group_statements(node)
            self.visit(groups, ctx=ctx)

    def visit_AxisBound(self, node: oir.AxisBound, axis_start: str, axis_end: str) -> str:
        if node.level == common.LevelMarker.START:
            return f"({axis_start}) + ({node.offset})"

        return f"({axis_end}) + ({node.offset})"

    def visit_Interval(
        self, node: oir.Interval, loop_order: common.LoopOrder, axis_start: str, axis_end: str
    ) -> tir.Bounds:
        start = self.visit(node.start, axis_start=axis_start, axis_end=axis_end)
        end = self.visit(node.end, axis_start=axis_start, axis_end=axis_end)

        if loop_order == common.LoopOrder.BACKWARD:
            return tir.Bounds(start=f"{end} - 1", end=start)

        return tir.Bounds(start=start, end=end)

    def _vertical_loop_schedule(self) -> dtypes.ScheduleType:
        """
        Defines the vertical loop schedule.

        Current strategy is to
          - keep the vertical loop on the host for both, CPU and GPU targets
          - and run it in parallel on CPU and sequential on GPU.
        """
        if self._device_type == dtypes.DeviceType.GPU:
            return dtypes.ScheduleType.Sequential

        return DEFAULT_MAP_SCHEDULE[self._device_type]

    def visit_VerticalLoopSection(
        self, node: oir.VerticalLoopSection, ctx: tir.Context, loop_order: common.LoopOrder
    ) -> None:
        bounds = self.visit(
            node.interval,
            loop_order=loop_order,
            axis_start="0",
            axis_end=tir.Axis.K.domain_dace_symbol(),
        )

        loop = tir.VerticalLoop(
            loop_order=loop_order,
            bounds_k=bounds,
            schedule=self._vertical_loop_schedule(),
            children=[],
            parent=ctx.current_scope,
        )

        with loop.scope(ctx):
            self.visit(node.horizontal_executions, ctx=ctx)

    def visit_VerticalLoop(self, node: oir.VerticalLoop, ctx: tir.Context) -> None:
        if node.caches:
            raise NotImplementedError("Caches are not supported in this prototype.")

        self.visit(node.sections, ctx=ctx, loop_order=node.loop_order)

    def visit_Stencil(self, node: oir.Stencil) -> tir.TreeRoot:
        # setup the descriptor repository
        containers: dict[str, data.Data] = {}
        dimensions: dict[str, tuple[bool, bool, bool]] = {}
        symbols: tir.SymbolDict = {}
        shift: dict[str, dict[tir.Axis, int]] = {}  # dict of field_name -> (dict of axis -> shift)

        # this is ij blocks = horizontal execution
        field_extents, block_extents = oir_utils.compute_extents(
            node,
            centered_extent=True,
        )
        # When determining the shape of the array, we have to look at the field extents at large.
        # GT4Py tries to give a precise measure by looking at the horizontal restriction and reduce
        # the extent to the only grid points inside the mask. DaCe requires the real size of the
        # data, hence the call with ignore_horizontal_mask=True.
        field_without_mask_extents = oir_utils.compute_fields_extents(
            node,
            centered_extent=True,
            ignore_horizontal_mask=True,
        )

        missing_api_parameters: list[str] = [p.name for p in self._api_signature]
        for param in node.params:
            missing_api_parameters.remove(param.name)
            if isinstance(param, oir.ScalarDecl):
                containers[param.name] = data.Scalar(
                    dtype=utils.data_type_to_dace_typeclass(param.dtype),
                    debuginfo=utils.get_dace_debuginfo(param),
                )
                continue

            if isinstance(param, oir.FieldDecl):
                field_extent = field_extents[param.name]
                k_bound = self._k_bounds[param.name]
                shift[param.name] = {
                    tir.Axis.I: -field_extent[0][0],
                    tir.Axis.J: -field_extent[1][0],
                    tir.Axis.K: max(k_bound[0], 0),
                }
                containers[param.name] = data.Array(
                    dtype=utils.data_type_to_dace_typeclass(param.dtype),
                    shape=get_dace_shape(
                        param,
                        field_without_mask_extents[param.name],
                        k_bound,
                        symbols,
                    ),
                    strides=get_dace_strides(param, symbols),
                    storage=DEFAULT_STORAGE_TYPE[self._device_type],
                    debuginfo=utils.get_dace_debuginfo(param),
                )
                dimensions[param.name] = param.dimensions
                continue

            raise ValueError(f"Unexpected parameter type {type(param)}.")

        for field in node.declarations:
            field_extent = field_extents[field.name]
            k_bound = self._k_bounds[field.name]
            shift[field.name] = {
                tir.Axis.I: -field_extent[0][0],
                tir.Axis.J: -field_extent[1][0],
                tir.Axis.K: max(k_bound[0], 0),
            }
            # TODO / Dev Note: Persistent memory is an overkill here - we should scope
            # the temporary as close to the tasklets as we can, but any lifetime lower
            # than persistent will yield issues with memory leaks.
            containers[field.name] = data.Array(
                dtype=utils.data_type_to_dace_typeclass(field.dtype),
                shape=get_dace_shape(field, field_extent, k_bound, symbols),
                strides=get_dace_strides(field, symbols),
                transient=True,
                lifetime=dtypes.AllocationLifetime.Persistent,
                storage=DEFAULT_STORAGE_TYPE[self._device_type],
                debuginfo=utils.get_dace_debuginfo(field),
            )
            dimensions[field.name] = field.dimensions

        tree = tir.TreeRoot(
            name=node.name,
            containers=containers,
            dimensions=dimensions,
            shift=shift,
            symbols=symbols,
            children=[],
            parent=None,
        )

        ctx = tir.Context(
            root=tree,
            current_scope=tree,
            field_extents=field_extents,
            block_extents=block_extents,
        )

        self.visit(node.vertical_loops, ctx=ctx)

        return ctx.root

    # Visit expressions for condition code in ControlFlow
    def visit_Cast(self, node: oir.Cast, **kwargs: Any) -> str:
        dtype = utils.data_type_to_dace_typeclass(node.dtype)
        expression = self.visit(node.expr, **kwargs)

        return f"{dtype}({expression})"

    def visit_CartesianOffset(
        self, node: common.CartesianOffset, field: oir.FieldAccess, ctx: tir.Context, **_kwargs: Any
    ) -> str:
        shift = ctx.root.shift[field.name]
        indices: list[str] = []

        offset_dict = node.to_dict()
        for index, axis in enumerate(tir.Axis.dims_3d()):
            if ctx.root.dimensions[field.name][index]:
                shift_str = f" + {shift[axis]}" if shift[axis] != 0 else ""
                indices.append(
                    f"{axis.iteration_symbol()}{shift_str} + {offset_dict[axis.lower()]}"
                )

        return ", ".join(indices)

    def visit_VariableKOffset(
        self, node: oir.VariableKOffset, field: oir.FieldAccess, ctx: tir.Context, **kwargs: Any
    ) -> str:
        shift = ctx.root.shift[field.name]
        i_shift = f" + {shift[tir.Axis.I]}" if shift[tir.Axis.I] != 0 else ""
        j_shift = f" + {shift[tir.Axis.J]}" if shift[tir.Axis.J] != 0 else ""
        k_shift = f" + {shift[tir.Axis.K]}" if shift[tir.Axis.K] != 0 else ""

        return (
            f"{tir.Axis.I.iteration_symbol()}{i_shift}, "
            f"{tir.Axis.J.iteration_symbol()}{j_shift}, "
            f"{tir.Axis.K.iteration_symbol()}{k_shift} + {self.visit(node.k, ctx=ctx, **kwargs)}"
        )

    def visit_ScalarAccess(self, node: oir.ScalarAccess, **kwargs: Any) -> str:
        return f"{node.name}"

    def visit_FieldAccess(self, node: oir.FieldAccess, **kwargs: Any) -> str:
        if node.data_index:
            raise NotImplementedError("Data dimensions aren't supported yet.")

        if "field" in kwargs:
            kwargs.pop("field")

        field_name = node.name
        offsets = self.visit(node.offset, field=node, **kwargs)

        return f"{field_name}[{offsets}]"

    def visit_Literal(self, node: oir.Literal, **kwargs: Any) -> str:
        if type(node.value) is str:
            # Note: isinstance(node.value, str) also matches the string enum `BuiltInLiteral`
            # which we don't want to match because it returns lower-case `true`, which isn't
            # defined in (python) tasklet code.
            return node.value

        return self.visit(node.value, **kwargs)

    def visit_BuiltInLiteral(self, node: common.BuiltInLiteral, **kwargs: Any) -> str:
        if node == common.BuiltInLiteral.TRUE:
            return "True"

        if node == common.BuiltInLiteral.FALSE:
            return "False"

        raise NotImplementedError(f"Not implemented BuiltInLiteral '{node}' encountered.")

    def visit_UnaryOp(self, node: oir.UnaryOp, **kwargs: Any) -> str:
        expression = self.visit(node.expr, **kwargs)

        return f"{node.op}({expression})"

    def visit_BinaryOp(self, node: oir.BinaryOp, **kwargs: Any) -> str:
        left = self.visit(node.left, **kwargs)
        right = self.visit(node.right, **kwargs)

        return f"({left} {node.op.value} {right})"

    def visit_TernaryOp(self, node: oir.TernaryOp, **kwargs: Any) -> str:
        condition = self.visit(node.cond, **kwargs)
        if_code = self.visit(node.true_expr, **kwargs)
        else_code = self.visit(node.false_expr, **kwargs)

        return f"({if_code} if {condition} else {else_code})"

    # visitors that should _not_ be called

    def visit_Decl(self, node: oir.Decl, **kwargs: Any) -> None:
        raise RuntimeError("visit_Decl should not be called")

    def visit_FieldDecl(self, node: oir.FieldDecl, **kwargs: Any) -> None:
        raise RuntimeError("visit_FieldDecl should not be called")

    def visit_LocalScalar(self, node: oir.LocalScalar, **kwargs: Any) -> None:
        raise RuntimeError("visit_LocalScalar should not be called")


def get_dace_shape(
    field: oir.FieldDecl,
    extent: definitions.Extent,
    k_bound: tuple[int, int],
    symbols: tir.SymbolDict,
) -> list[symbolic.symbol]:
    shape = []
    for index, axis in enumerate(tir.Axis.dims_3d()):
        if field.dimensions[index]:
            symbol = axis.domain_dace_symbol()
            symbols[axis.domain_symbol()] = dtypes.int32

            if axis == tir.Axis.I:
                i_padding = extent[0][1] - extent[0][0]
                if i_padding != 0:
                    shape.append(symbol + i_padding)
                    continue

            if axis == tir.Axis.J:
                j_padding = extent[1][1] - extent[1][0]
                if j_padding != 0:
                    shape.append(symbol + j_padding)
                    continue

            if axis == tir.Axis.K:
                k_padding = max(k_bound[0], 0) + max(k_bound[1], 0)
                if k_padding != 0:
                    shape.append(symbol + k_padding)
                    continue

            shape.append(symbol)

    shape.extend([d for d in field.data_dims])
    return shape


def get_dace_strides(field: oir.FieldDecl, symbols: tir.SymbolDict) -> list[symbolic.symbol]:
    dimension_strings = [d for i, d in enumerate("IJK") if field.dimensions[i]]
    data_dimension_strings = [f"d{ddim}" for ddim in range(len(field.data_dims))]

    strides = []
    for dim in dimension_strings + data_dimension_strings:
        stride = f"__{field.name}_{dim}_stride"
        symbol = symbolic.pystr_to_symbolic(stride)
        symbols[stride] = dtypes.int32
        strides.append(symbol)
    return strides
